/* Barcode scanning (wand, flexible timing)
 *
 * Specifically, this module has the following requirements on barcodes:
 *  - symbology: interleaved 2 of 5
 *  - 4 or 6 (numeric) characters; default is 4
 *  - no checksum, but you may use the last digit as a checksum
 */
#include <rational>

#if !defined barcode_digits
  const barcode_digits = 4
#endif

barcode_init(sample_pin, timeout=400, power_pin=-1)
    {
    if (power_pin >= 0)
        {
        configiopin power_pin, Output
        setiopin power_pin, 1
        }

    configiopin sample_pin, Sample, timeout
    }

@sample(const Fixed:stamps[], numsamples)
    {
    /* There are two kinds of wand scanners: inverted and non-inverted.
     * The difference is the value a black bar: it is either 0 or 1.
     * For this routine, the difference is that for the non-inverted
     * scanner, there is an initial time stamp with the value 0.000
     * in the array.
     */
    new base = 0
    if (stamps[0] == 0.0)
        base = 1

    /* Calculate the average width of the narrow bar/gap. This value
     * may be refined later, in the first scan through all samples.
     */
    new Fixed: narrow = stamps[base + 3] / 4
    new Fixed: narrow_min       /* these will be calculated later */
    new Fixed: narrow_max
    new Fixed: wide_max

    /* We can calculate the number of samples needed for a number of digits:
     * the start en termination codes are two bars each and there are 5
     * bars for each two digits. For 6 digits (3 pairs), this means that
     * the number of bars = 6/2 * 5 + 2 + 2. The number of samples is the
     * double of this: 38, but the initial edge is stripped from this.
     */
    const expectsamples = barcode_digits * 5 + 8 - 1
    if (numsamples - base < expectsamples)
        {
        barcode -1
        return
        }
    if (numsamples - base > expectsamples)
        numsamples = expectsamples + base       /* clamp to max. */

    /* Scan through the list and check for very short or very long
     * bars/gaps. At the same time, collect the narrow and wide
     * bars/gaps in a separate array.
     */
    new samples[expectsamples]
    new top
    new Fixed: prev = 0.0
    new Fixed: dif
    for (top = base; top < numsamples; top++)
        {
        /* calculate the boundary criterions */
        narrow_min = narrow * 0.5
        narrow_max = narrow * 1.5
        wide_max = narrow * 4.0
        /* see whether a sample width is wide, narrow, or out of range */
        dif = stamps[top] - prev
        prev = stamps[top]
        if (dif < narrow_min || dif > wide_max)
            {
            /* bar/gap is out-of-range (too narrow or too wide) */
            barcode -1
            return
            }
        if (dif >= narrow_max)
            samples[top - base] = 1
        else
            {
            samples[top - base] = 0
            /* Adjust the value for the "narrow bar/gap", so that
             * accelerations and decelerations in scanning speed are
             * handled adaptively.
             */
            narrow = (narrow + dif) / 2
            }
        }
    /* now we also have a clean base (starting from 0) */
    numsamples -= base

    /* check the start and end code */
    if (samples[0] || samples[1] || samples[2] || samples[3])
        {
        barcode -1
        return
        }

    if (samples[barcode_digits * 5 + 4] == 0
        || samples[barcode_digits * 5 + 5]
        || samples[barcode_digits * 5 + 6])
        {
        barcode -1
        return
        }

    /* if you do not use a checksum digit, one sould at least check the
     * parity bit; here we define the required table
     */                /* 0  1  2  3  4  5  6  7  8  9 */
    static parity[10] = { 0, 1, 1, 0, 1, 0, 0, 1, 0, 0 }

    /* get the value
     * Interleaved 2 of 5 is a almost a binary code where a narrow bar/gap
     * stands for a zero and a wide bar/gap for a 1. It uses four bars/gaps
     * for the value and the fifth as a parity check. In a normal binary code,
     * the bit positions are powers of 2; when set, the bits contribute 1, 2,
     * 4 and 8 to the value. In interleaved 2 of 5, the last bit contributes
     * 7 to the value (not eight). This has the consequence that some values
     * (such as 7 itself) can be coded in two different ways. However, there
     * is an additional requirement that exactly two bars/gaps out of 5 are
     * wide (i.e. 1). A side effect of this is that zero is also a special
     * case: it is encoded as 11.
     */
    new value = 0
    new digit
    base = 4    /* skip start code */
    while (base < expectsamples - 7)
        {
        digit = samples[base] + 2 * samples[base + 2]
                + 4 * samples[base + 4] + 7 * samples[base + 6]
        if (digit == 11)
            digit = 0
        if (samples[base + 8] != parity[digit])
            {
            barcode -1
            return
            }
        value = 10 * value + digit

        digit = samples[base + 1] + 2 * samples[base + 3]
                + 4 * samples[base + 5] + 7 * samples[base + 7]
        if (digit == 11)
            digit = 0
        if (samples[base + 8] != parity[digit])
            {
            barcode -1
            return
            }
        value = 10 * value + digit

        base += 10
        }

    /* all check passed, call the function that acts on the barcode */
    barcode value
    }
